from django.test import TestCase
from django.views.generic import View
from django.views.generic.detail import SingleObjectMixin

from ai_django_core.tests.mixins import RequestProviderMixin
from ai_django_core.view_layer.views import ToggleView


class ToggleViewTest(RequestProviderMixin, TestCase):

    def test_http_method_set_correctly(self):
        self.assertEqual(ToggleView.http_method_names, ('post',))

    def test_post_raises_not_implemented_error(self):
        with self.assertRaises(NotImplementedError):
            view = ToggleView()
            view.post(request=self.get_request())

    def test_class_inherits_from_single_object_mixin(self):
        self.assertTrue(issubclass(ToggleView, SingleObjectMixin))

    def test_class_inherits_from_generic_view(self):
        self.assertTrue(issubclass(ToggleView, View))



import json
import os
import subprocess
import sys
from typing import Optional

import httpx


class CoverageService:
    """
    Class to be used in the gitlab-ci to ensure pipeline test coverage is not dropping on commit
    """

    def __init__(self) -> None:
        super().__init__()

        # Get ENV variables
        self.current_pipeline_id: int = int(os.environ.get('CI_PIPELINE_ID'))
        self.base_api_url: str = os.environ.get('CI_API_V4_URL')
        self.current_branch: str = os.environ.get('CI_COMMIT_REF_NAME')
        self.token: str = os.environ.get('GITLAB_CI_COVERAGE_PIPELINE_TOKEN')
        self.project_id: int = int(os.environ.get('CI_PROJECT_ID'))
        self.job_name: str = os.environ.get('CI_COVERAGE_JOB_NAME', '')
        self.target_branch: str = os.environ.get('GITLAB_CI_COVERAGE_TARGET_BRANCH')
        self.pipelines_url = (
            f'{self.base_api_url}/projects/{self.project_id}/pipelines?ref={self.target_branch}&status=success'
        )
        self.pipelines_url_with_token = f'{self.pipelines_url}&private_token={self.token}'

    def get_latest_target_branch_commit_sha(self) -> str:
        """
        Get the latest commit which is in the current branch and the target compare branch.
        """
        result = subprocess.run(
            ['git', 'merge-base', '--fork-point', f'origin/{self.target_branch}'], stdout=subprocess.PIPE
        )
        return result.stdout.decode("utf-8").strip()

    def get_pipeline_id_by_commit_sha(self, sha: str) -> Optional[int]:
        pipeline_url = f'{self.pipelines_url_with_token}&sha={sha}'
        response = httpx.get(pipeline_url)
        status_code = response.status_code

        if status_code == 200:
            pipelines = json.loads(response.content)
            if pipelines and len(pipelines) > 0:
                return pipelines[0].get('id', None)
            else:
                print('\n### ERROR: No pipelines found for SHA1 ###\n')
                print(f'Pipeline URL: {self.pipelines_url}&sha={sha}')
                print(response.content)
        return None

    def get_coverage_from_pipeline(self, pipeline_id: int, job_name: str) -> (float, float):
        """
        Get coverage from given pipeline (by id)
        """
        jobs_url = f'{self.base_api_url}/projects/{self.project_id}/pipelines/{pipeline_id}/jobs'
        jobs_with_token_url = f'{jobs_url}?private_token={self.token}'

        print(f'Jobs URL: {jobs_url}')
        jobs_response = httpx.get(jobs_with_token_url)
        jobs_status_code = jobs_response.status_code

        if jobs_status_code != 200:
            raise ConnectionError(f'Call to jobs api endpoint failed with status code {jobs_status_code}')

        jobs = json.loads(jobs_response.content)
        coverages = {job['name']: float(job['coverage']) for job in jobs if job.get('coverage')}

        pipeline_url = (
            f'{self.base_api_url}/projects/{self.project_id}/pipelines/{pipeline_id}'
        )
        pipeline_with_token_url = f'{pipeline_url}?private_token={self.token}'
        print(f'Pipeline URL: {pipeline_url}')
        pipeline_response = httpx.get(pipeline_with_token_url)
        pipeline_status_code = pipeline_response.status_code

        if pipeline_status_code != 200:
            raise ConnectionError(f'Call to pipeline api endpoint failed with status code {pipeline_status_code}')

        pipeline = json.loads(pipeline_response.content)
        coverages_total = float(pipeline['coverage'] if pipeline['coverage'] else 0.0)

        if job_name == '':
            return coverages_total, coverages_total
        return coverages.get(job_name) if coverages.get(job_name) else 0.0, coverages_total

    @staticmethod
    def color_text(sign: int, prefix: str, target: float, current: float, diff: float):
        """
        function to return colored in text according to the template:
        "{Total|Job} coverage {change_text} from {target}% to {current}% (Diff: {diff}%)"
        Red color: coverage dropped
        White/No color: coverage unchanged
        Green color: coverage climbed
        :param sign: numeric value of the coverage diff sign
        :param prefix: text prefix (i.e.: Total coverage, Job Coverage)
        :param target: target percentage
        :param current: current percentage
        :param diff: difference between current and target percentage
        :return: fully assembled and colored summary text
        """
        change = {
            -1: {'text': 'dropped', 'color': '\033[91m'},
            0: {'text': 'unchanged', 'color': ''},
            1: {'text': 'climbed', 'color': '\033[92m'},
        }
        return (
            f'{change[sign]["color"]} {prefix} {change[sign]["text"]} '
            f'from {target:2.2f}% to {current:2.2f}% (Diff: {diff:2.2f}%).\033[0m'
        )

    def process(self):
        """
        Compare coverage from target branch (latest develop) with the current one.
        At first, we try to get a successfully finished pipeline for the "self.target_branch" (usually "develop")
        """
        print('\n###########################################################################\n')
        print('DEBUG INFO:')

        # Get the latest commit SHA which is also in develop
        commit_sha = self.get_latest_target_branch_commit_sha()

        # Try to find the latest successful pipeline for "TARGET_BRANCH" where our SHA was in
        print('Trying base branch for comparison.')
        target_pipeline_id = None
        if commit_sha:
            print(f'Found latest target branch commit SHA "{commit_sha}".')
            target_pipeline_id = self.get_pipeline_id_by_commit_sha(commit_sha)
            print(f'Target branch pipeline ID identified: {target_pipeline_id}.')

        # Get target pipeline id (from develop branch) if we were not successful the first time
        if not target_pipeline_id:
            print("Didn't work. Using default branch for comparison.")
            response = httpx.get(self.pipelines_url_with_token)
            status_code = response.status_code
            print(f'Pipelines-URL: {self.pipelines_url}')

            # Ensure call did not go sideways
            if status_code != 200:
                raise ConnectionError(f'Call to global pipeline api endpoint failed with status code {status_code}')

            # Read target pipeline ID from content
            target_pipeline_id = json.loads(response.content)[0]['id']
            print(f'Default branch pipeline ID identified: {target_pipeline_id}.')

        # Get coverage from target pipeline
        print(f'Target Pipeline ID: {target_pipeline_id}')
        target_job_coverage, target_total_coverage = self.get_coverage_from_pipeline(target_pipeline_id, self.job_name)

        # Get coverage from this pipeline
        print(f'Current pipeline ID: {self.current_pipeline_id}')
        current_job_coverage, current_total_coverage = self.get_coverage_from_pipeline(
            self.current_pipeline_id, self.job_name
        )

        # numeric value of the coverage diff sign
        sign_job_coverage = (current_job_coverage > target_job_coverage) - (current_job_coverage < target_job_coverage)
        sign_total_coverage = (current_total_coverage > target_total_coverage) - (
                current_total_coverage < target_total_coverage
        )

        # difference between current and target coverage
        diff_job_coverage = current_job_coverage - target_job_coverage
        diff_total_coverage = current_total_coverage - target_total_coverage

        coverage = {
            'total': {
                'target': target_total_coverage,
                'current': current_total_coverage,
                'sign': sign_total_coverage,
                'diff': diff_total_coverage,
                'prefix': 'Total coverage',
            },
            'job': {
                'target': target_job_coverage,
                'current': current_job_coverage,
                'sign': sign_job_coverage,
                'diff': diff_job_coverage,
                'prefix': 'Job coverage',
            },
        }

        print('\n###########################################################################\n')

        # Print results
        print(
            self.color_text(
                coverage['total']['sign'],
                coverage['total']['prefix'],
                coverage['total']['target'],
                coverage['total']['current'],
                coverage['total']['diff'],
            )
        )
        print(
            self.color_text(
                coverage['job']['sign'],
                coverage['job']['prefix'],
                coverage['job']['target'],
                coverage['job']['current'],
                coverage['job']['diff'],
            )
        )
        if coverage['job']['sign'] == -1:
            sys.exit(1)



from django.contrib import admin
from django.urls import path, include
from rest_framework import routers

from testapp.api.urls import model_router

router = routers.DefaultRouter()
router.registry.extend(model_router.registry)

urlpatterns = [
    # django Admin
    path('admin/', admin.site.urls),

    # REST Viewsets
    path('api/v1/', include(router.urls)),
]



from django.contrib import admin
from django.contrib.auth.models import User
from django.test import TestCase

from ai_django_core.admin.model_admins.classes import ReadOnlyAdmin, EditableOnlyAdmin
from ai_django_core.tests.mixins import RequestProviderMixin
from testapp.models import MySingleSignalModel, MyMultipleSignalModel


class TestReadOnlyAdmin(ReadOnlyAdmin):
    pass


class TestEditableOnlyAdmin(EditableOnlyAdmin):
    pass


class AdminClassesTest(RequestProviderMixin, TestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.super_user = User.objects.create(username='super_user', is_superuser=True)

        admin.site.register(MySingleSignalModel, TestReadOnlyAdmin)
        admin.site.register(MyMultipleSignalModel, TestEditableOnlyAdmin)

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

        admin.site.unregister(MySingleSignalModel)
        admin.site.unregister(MyMultipleSignalModel)

    def test_read_only_admin_all_fields_readonly(self):
        obj = MySingleSignalModel(value=1)

        admin_class = TestReadOnlyAdmin(model=obj, admin_site=admin.site)
        readonly_fields = admin_class.get_readonly_fields(request=self.get_request(), obj=obj)

        self.assertEqual(len(readonly_fields), 2)
        self.assertIn('id', readonly_fields)
        self.assertIn('value', readonly_fields)

    def test_read_only_admin_no_change_permissions(self):
        admin_class = TestReadOnlyAdmin(model=MySingleSignalModel, admin_site=admin.site)

        request = self.get_request(self.super_user)

        self.assertFalse(admin_class.has_add_permission(request))
        self.assertFalse(admin_class.has_change_permission(request))
        self.assertFalse(admin_class.has_delete_permission(request))

    def test_editable_only_admin_delete_action_removed(self):
        obj = MyMultipleSignalModel(value=1)
        admin_class = TestEditableOnlyAdmin(model=obj, admin_site=admin.site)

        request = self.get_request(self.super_user)
        actions = admin_class.get_actions(request=request)

        self.assertNotIn('delete_selected', actions)

    def test_editable_only_admin_no_change_permissions(self):
        admin_class = TestEditableOnlyAdmin(model=MyMultipleSignalModel, admin_site=admin.site)

        request = self.get_request(self.super_user)

        self.assertTrue(admin_class.has_change_permission(request))

        self.assertFalse(admin_class.has_add_permission(request))
        self.assertFalse(admin_class.has_delete_permission(request))



from math import ceil


def round_to_decimal(value, precision: float = 0.5) -> float:
    """
    Helper function to round a given value to a specific precision, for example *.5
    So 5.4 will be rounded to 5.5
    """
    return round(precision * round(float(value) / precision), 1)


def round_up_decimal(value, precision: float = 0.5) -> float:
    """
    Helper function to round a given value up a specific precision, for example *.5
    So 5.4 will be rounded to 5.5 and 5.6 to 6.0
    """
    return ceil(value * (1 / precision)) / (1 / precision)



from django.core import mail
from django.db.models import signals
from django.test import TestCase

from ai_django_core.context_manager import TempDisconnectSignal
from testapp.models import MySingleSignalModel, increase_value_no_dispatch_uid, increase_value_with_dispatch_uid, \
    MyMultipleSignalModel, send_email


class TempDisconnectSignalTest(TestCase):

    def test_single_signal_executed_regular(self):
        obj = MySingleSignalModel.objects.create()
        self.assertEqual(obj.value, 1)

    def test_single_signal_not_executed(self):
        kwargs = {
            'signal': signals.pre_save,
            'receiver': increase_value_no_dispatch_uid,
            'sender': MySingleSignalModel,
        }

        with TempDisconnectSignal(**kwargs):
            obj = MySingleSignalModel.objects.create()

        self.assertEqual(obj.value, 0)

    def test_multiple_signals_not_executed(self):
        kwargs_pre = {
            'signal': signals.pre_save,
            'receiver': increase_value_with_dispatch_uid,
            'sender': MyMultipleSignalModel,
            'dispatch_uid': 'test.mysinglesignalmodel.increase_value_with_uuid',
        }
        kwargs_post = {
            'signal': signals.post_save,
            'receiver': send_email,
            'sender': MyMultipleSignalModel,
        }

        with TempDisconnectSignal(**kwargs_pre):
            with TempDisconnectSignal(**kwargs_post):
                obj = MyMultipleSignalModel.objects.create()

        outbox = len(mail.outbox)

        self.assertEqual(obj.value, 0)
        self.assertEqual(outbox, 0)

    def test_multiple_signals_one_still_active(self):
        kwargs_pre = {
            'signal': signals.pre_save,
            'receiver': increase_value_with_dispatch_uid,
            'sender': MyMultipleSignalModel,
            'dispatch_uid': 'test.mysinglesignalmodel.increase_value_with_uuid',
        }

        with TempDisconnectSignal(**kwargs_pre):
            obj = MyMultipleSignalModel.objects.create()

        outbox = len(mail.outbox)

        self.assertEqual(obj.value, 0)
        self.assertEqual(outbox, 1)



def strip_sensitive_data_from_sentry_event(event, hint):
    """
    Helper method to strip sensitive user data from default sentry event when "send_default_pii" is set to True.
    All user-related data except the internal user id will be removed.
    Variable "hint" contains information about the error itself which we don't need here.
    Requires "sentry-sdk>=1.5.0" to work.
    """
    try:
        del event['user']['username']
    except KeyError:
        pass
    try:
        del event['user']['email']
    except KeyError:
        pass
    try:
        del event['user']['ip_address']
    except KeyError:
        pass
    return event



from typing import List

from django.core import mail
from django.test import TestCase
import warnings


class EmailTestService:
    _outbox = None

    def _ensure_outbox_is_loaded(self):
        """
        Ensures that the outbox attribute is set
        """
        if self._outbox is None:
            self.reload()

    def reload(self):
        """
        Loads the current _outbox inside an attribute of this class
        """
        self._outbox = mail.outbox

    def empty(self):
        """
        Empties the current outbox
        :return:
        """
        mail.outbox = []
        self.reload()

    def filter(self, to=None, cc=None, bcc=None, subject=None):
        """
        Searches in the _outbox for emails matching either to and/or subject.
        Returns a list of email objects
        :param to: str
        :param cc: str
        :param bcc: str
        :param subject: str
        :return: EmailTestService
        """
        # Ensure that outbox is up-to-date
        self.reload()

        match_list = []
        for email in self._outbox:
            # Check conditions
            match = True
            if to and to not in email.to:
                match = False
            if cc and cc not in email.cc:
                match = False
            if bcc and bcc not in email.bcc:
                match = False
            if subject and not email.subject == subject:
                match = False
            if not to and not cc and not bcc and not subject:
                raise ValueError('EmailTestService finder called without parameters')

            # Add email if all set conditions are valid
            if match:
                match_list.append(email)

        return EmailTestServiceQuerySet(matching_list=match_list)

    def all(self):
        """
        Loads all mails from the outbox inside the matching list
        :return:
        """
        # Ensure that outbox is up-to-date
        self.reload()

        # Load data to matching list
        match_list = []
        for email in self._outbox:
            match_list.append(email)

        return EmailTestServiceQuerySet(matching_list=match_list)


class EmailTestServiceMail(mail.EmailMultiAlternatives):
    """
    Wrapper around a Django EmailMultiAlternatives object with some helper functions to write clean assertions.
    """
    _testcase = TestCase()  # Hacky way to get access to TestCase.assert* methods without deriving from TestCase

    def _get_html_content(self):
        """
        Ensure we just have found one element and then return HTML part of the email
        :return: str
        """
        # Search for string
        if len(self.alternatives) > 0:
            return self.alternatives[0][0]
        return None

    def _get_txt_content(self):
        """
        Ensure we just have found one element and then return text part of the email
        :return: str
        """
        # Search for string
        return self.body

    def assert_subject(self, subject, msg=None):
        """
        Searches in a given email inside the HTML AND TXT part for a given string
        :param subject: str
        :param msg: str
        """

        # Assert expected subject is equal to the generated one
        self._testcase.assertEqual(subject, self.subject, msg=msg)

    def assert_body_contains(self, search_str, msg=None):
        """
        Searches in a given email inside the HTML AND TXT part for a given string
        :param search_str: str
        :param msg: str
        """

        # Assert string is contained in TXT part
        self._testcase.assertIn(search_str, self._get_txt_content(), msg=msg)
        # Assert string is contained in HTML part (if HTML part is set)
        html_content = self._get_html_content()
        if html_content is not None:
            self._testcase.assertIn(search_str, html_content, msg=msg)

    def assert_body_contains_not(self, search_str, msg=None):
        """
        Searches in a given email inside the HTML AND TXT part for a given string
        :param search_str: str
        :param msg: str
        """

        # Assert string is contained in HTML part
        self._testcase.assertNotIn(search_str, self._get_html_content(), msg=msg)
        # Assert string is contained in TXT part
        self._testcase.assertNotIn(search_str, self._get_txt_content(), msg=msg)

    def assert_to_contains(self, *emails: List[str]):
        for email in emails:
            self._testcase.assertIn(email, self.to)


class EmailTestServiceQuerySet(TestCase):
    _match_list = None

    def __init__(self, matching_list=None):
        super().__init__()
        self._match_list = matching_list
        for email in self._match_list or []:
            # Change the class of every EmailMutliAlternative instance, so that it points to
            # our subclass, which has some additional assertion-methods.
            email.__class__ = EmailTestServiceMail

    def _get_html_content(self):
        """
        Ensure we just have found one element and then return HTML part of the email
        :return: str
        """
        self._validate_lookup_cache_contains_one_element()
        # Search for string
        if len(self[0].alternatives) > 0:
            return self[0].alternatives[0][0]
        return None

    def _get_txt_content(self):
        """
        Ensure we just have found one element and then return text part of the email
        :return: str
        """
        # Search for string
        self._validate_lookup_cache_contains_one_element()
        return self[0].body

    def _validate_lookup_cache_contains_one_element(self):
        """
        Ensures that in the cached lookup is exactly one element. Needed for full-text-search.
        """
        if self.count() > 1:
            raise RuntimeError('Current lookup has more than one email object and is thus ambiguous.')
        elif self.count() == 0:
            raise RuntimeError('Current lookup has zero matches so lookup does not make sense.')

    def _ensure_matching_list_was_populated(self):
        """
        Make sure that we queried at least once before working with the results
        """
        if self._match_list is None:
            raise RuntimeError('Counting of matches called without previous query. '
                               'Please call filter() or all() first.')

    def one(self):
        """
        Checks if the previous query returned exactly one element
        :return: bool
        """
        return self.count() == 1

    def count(self):
        """
        Returns the number of matches found by a previous call of `find()`
        :return: int
        """
        # Ensure is was queried before using results
        self._ensure_matching_list_was_populated()

        # Count matches
        return len(self)

    def first(self):
        """
        Returns the first found element
        :return: EmailMultiAlternatives
        """
        # Ensure is was queried before using results
        self._ensure_matching_list_was_populated()
        return self[0] if self.count() > 0 else False

    def last(self):
        """
        Returns the last found element
        :return: EmailMultiAlternatives
        """
        # Ensure is was queried before using results
        self._ensure_matching_list_was_populated()
        return self[-1] if self.count() > 0 else False

    def assert_one(self, msg=None):
        """
        Makes an assertion to make sure queried element exists exactly once
        :param msg: str
        :return:
        """
        self.assertEqual(self.one(), True, msg=msg)

    def assert_quantity(self, target_quantity, msg=None):
        """
        Makes an assertion to make sure that amount of queried mails are equal to `target_quantity`
        :param target_quantity: int
        :param msg: str
        :return:
        """
        self.assertEqual(self.count(), target_quantity, msg=msg)

    def assert_subject(self, subject, msg=None):
        """
        Searches in a given email inside the HTML AND TXT part for a given string
        :param subject: str
        :param msg: str
        """
        warnings.warn('EmailTestServiceQuerySet.assert_subject is deprecated. Use queryset[0].assert_subject instead.')
        # Ensure we just have found one element
        self._validate_lookup_cache_contains_one_element()
        self[0].assert_subject(subject, msg)

    def assert_body_contains(self, search_str, msg=None):
        """
        Searches in a given email inside the HTML AND TXT part for a given string
        :param search_str: str
        :param msg: str
        """
        warnings.warn('EmailTestServiceQuerySet.assert_body_contains is deprecated. '
                      'Use queryset[0].assert_body_contains instead.')
        # Ensure we just have found one element
        self._validate_lookup_cache_contains_one_element()
        self[0].assert_body_contains(search_str, msg)

    def assert_body_contains_not(self, search_str, msg=None):
        """
        Searches in a given email inside the HTML AND TXT part for a given string
        :param search_str: str
        :param msg: str
        """
        warnings.warn('EmailTestServiceQuerySet.assert_body_contains is deprecated. '
                      'Use queryset[0].assert_body_contains_not instead.')
        # Ensure we just have found one element
        self._validate_lookup_cache_contains_one_element()
        self[0].assert_body_contains_not(search_str, msg)

    def __getitem__(self, item):
        return self._match_list.__getitem__(item)

    def __len__(self):
        return self._match_list.__len__()



from django import template

register = template.Library()


@register.simple_tag
def dict_key_lookup(the_dict, key):
    """
    Checks if the given key exists in given dict

    :param the_dict:
    :param key:
    :return: str
    """
    return the_dict.get(key, '')


@register.filter(name='label')
def label(value):
    return value.field.__class__.__name__
